/*
 * Copyright (c) 2011, Make, Hack, Void Inc
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of the Make, Hack, Void nor the
 *    names of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MAKE, HACK, VOID BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* Generate a 7V output using a boost circuit fed from VCC
 *
 * Uses Timer1 to generate a PWM signal to drive the transistor.
 * The transistor should be connected to OC1B, which is Arduino pin 12 on the Mega
 * and Arduino pin 10 on others
 *
 * See http://www.ladyada.net/library/diyboostcalc.html for the schematic & calculator
 */

#include <util/delay.h>

// Bring in the MHV IO header
#include <MHV_io.h>

// Bring in the MHV 16 bit timer header
#include <MHV_Timer16.h>

// Bring in the MHV Voltage Regulator header
#include <MHV_VoltageRegulator.h>

// Bring in the MHV Serial header
#include <MHV_HardwareSerial.h>

// Bring in the MHV ADC header
#include <MHV_AD.h>

// Bring in the power management header
#include <avr/power.h>


MHV_HARDWARESERIAL_CREATE(serial, 0, 1, MHV_USART0, 115200);

// We will be operating using Timer1
#define OUTPUT_PIN MHV_PIN_TIMER_1_B
MHV_Timer16 vOutTimer(MHV_TIMER16_1);

#define TARGET_VOLTAGE 7.0

// This is only accurate to ~ 5%, you may want to measure it (on the AREF pin) to increase accuracy
#define REFERENCE_VOLTAGE 1.1

// The resistors in the resistor divider, the sum should not be greater than 10k
#define R1	7250.00
#define R2	800.00

// Which ADC channel to use
#define ADC_CHANNEL	MHV_AD_CHANNEL_0

MHV_VoltageRegulator regulator(MHV_VREG_MODE_BOOST, TARGET_VOLTAGE, REFERENCE_VOLTAGE,
		MHV_AD_REFERENCE_1V1, R2/(R2+R1), &vOutTimer, ADC_CHANNEL);

int main(void) {
	// Disable all peripherals and enable just what we need
	power_all_disable();
	power_usart0_enable();
	power_timer1_enable();

// Enable interrupts
	sei();

	mhv_ad_setPrescaler(MHV_AD_PRESCALER_128);
//	MHV_AD_ENABLE;

// Set the pin the transistor is connected to to output
	mhv_setOutput(OUTPUT_PIN);

// Set the PWM mode to FAST PWM
	vOutTimer.setMode(MHV_TIMER_16_PWM_FAST);

// Set the PWM prescaler to 1 (no prescaler)
	vOutTimer.setPrescaler(MHV_TIMER_PRESCALER_5_1);

/* Set the TOP value of the PWM timer - this defines the resolution &
 * frequency of the PWM output. With a 16MHz clock, this gives us a PWM
 * frequency of 4kHz
 */
	vOutTimer.setTop(4095);

/* Tell the pwmTimer to use output compare 2 for PWM output
 * It will turn off the output pin when the timer elapses, and turn it on
 * when it is reset
 */
	vOutTimer.connectOutput2(MHV_TIMER_CONNECT_CLEAR);

// Start the regulator
	regulator.enable();

// Update the voltage periodically
	for(;;) {
		regulator.regulate();

/* No floating point in the default printf library from AVR LibC
 * (it can be enabled with "-Wl,-u,vfprintf -lprintf_flt -lm" at the cost of memory)
 * so we will print the current voltage in milliVolts
 */
		uint16_t milliVolts = (uint16_t)(1000 * regulator.getVoltage());
		MHV_HARDWARESERIAL_DEBUG(serial, "Current Voltage is %u millivolts", milliVolts);

		_delay_ms(100);
	}

// Main must return an int, even though we never get here
	return 0;
}
